/*
 * indicator-network - user interface for connman
 * Copyright 2010 Canonical Ltd.
 *
 * Authors:
 * Kalle Valo <kalle.valo@canonical.com>
 *
 * This program is free software: you can redistribute it and/or modify it
 * under the terms of the GNU General Public License version 3, as published
 * by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranties of
 * MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
 * PURPOSE.  See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along
 * with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include "wireless-connect-dialog.h"

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <glib.h>
#include <glib-object.h>
#include <glib/gi18n.h>
#include <dbus/dbus-glib.h>
#include <gtk/gtk.h>
#include <locale.h>

#include "connman.h"
#include "connman-manager.h"
#include "connman-service.h"

G_DEFINE_TYPE(WirelessConnectDialog, wireless_connect_dialog, GTK_TYPE_DIALOG);

typedef struct _WirelessConnectDialogPriv WirelessConnectDialogPriv;

struct _WirelessConnectDialogPriv
{
  DBusGConnection *system_bus;
  DBusGProxy *connman_manager;
  DBusGProxy *service_proxy;
  GtkWidget *network_combo;
  GtkWidget *count_label;
  guint service_count;
  GtkListStore *store;
  DBusGProxyCall *services_call;
  DBusGProxyCall *connect_call;
};

#define WIRELESS_CONNECT_DIALOG_GET_PRIVATE(o)				\
  (G_TYPE_INSTANCE_GET_PRIVATE((o),					\
			       WIRELESS_CONNECT_DIALOG_TYPE,		\
			       WirelessConnectDialogPriv))

enum {
  COLUMN_PATH,
  COLUMN_NAME,
  N_COLUMNS,
};

static void wireless_connect_dialog_init(WirelessConnectDialog *self)
{
  WirelessConnectDialogPriv *priv = WIRELESS_CONNECT_DIALOG_GET_PRIVATE(self);
  GError *error = NULL;

  priv->system_bus = dbus_g_bus_get(DBUS_BUS_SYSTEM, &error);

  if (error) {
    g_critical("Unable to connect to the dbus system bus: %s",
	       error->message);
    g_error_free(error);
    return;
  }

  if (priv->system_bus == NULL) {
    g_critical("%s(): failed to get system bus", __func__);
    return;
  }

  priv->connman_manager = NULL;
  priv->service_proxy = NULL;
  priv->network_combo = NULL;
  priv->count_label = NULL;
  priv->service_count = 0;
  priv->store = NULL;
  priv->services_call = NULL;
  priv->connect_call = NULL;

  return;
}

static void wireless_connect_dialog_dispose(GObject *object)
{
  WirelessConnectDialog *self = WIRELESS_CONNECT_DIALOG(object);
  WirelessConnectDialogPriv *priv = WIRELESS_CONNECT_DIALOG_GET_PRIVATE(self);

  if (priv->service_proxy != NULL && priv->connect_call != NULL) {
    dbus_g_proxy_cancel_call(priv->service_proxy, priv->connect_call);
    priv->connect_call = NULL;
  }

  if (priv->connman_manager != NULL && priv->services_call != NULL) {
    dbus_g_proxy_cancel_call(priv->connman_manager, priv->services_call);
    priv->services_call = NULL;
  }

  if (priv->connman_manager != NULL) {
    g_object_unref(priv->connman_manager);
    priv->connman_manager = NULL;
  }

  if (priv->service_proxy != NULL) {
    g_object_unref(priv->service_proxy);
    priv->service_proxy = NULL;
  }

  G_OBJECT_CLASS(wireless_connect_dialog_parent_class)->dispose(object);
}

static void wireless_connect_dialog_finalize(GObject *object)
{
  G_OBJECT_CLASS(wireless_connect_dialog_parent_class)->finalize(object);
}

static void wireless_connect_dialog_class_init(WirelessConnectDialogClass *klass)
{
  GObjectClass *object_class = G_OBJECT_CLASS (klass);

  g_type_class_add_private(object_class, sizeof(WirelessConnectDialogPriv));

  object_class->dispose = wireless_connect_dialog_dispose;
  object_class->finalize = wireless_connect_dialog_finalize;

  g_assert(klass != NULL);
}

static void connect_cb(DBusGProxy *proxy, GError *error, gpointer user_data)
{
  WirelessConnectDialog *self;
  WirelessConnectDialogPriv *priv;


  self = WIRELESS_CONNECT_DIALOG(user_data);
  g_return_if_fail(self != NULL);
  g_return_if_fail(IS_WIRELESS_CONNECT_DIALOG(self));

  priv = WIRELESS_CONNECT_DIALOG_GET_PRIVATE(self);
  g_return_if_fail(priv != NULL);

  priv->connect_call = NULL;

  if (priv->service_proxy != NULL) {
    g_object_unref(priv->service_proxy);
    priv->service_proxy = NULL;
  }

  if (error) {
    /* FIXME: show warning to the user that the connection failed */
    g_warning("failed to connect service: %s",error->message);
    g_error_free(error);
    return;
  }
}

static void connect(WirelessConnectDialog *self, const gchar *path)
{
  WirelessConnectDialogPriv *priv = WIRELESS_CONNECT_DIALOG_GET_PRIVATE(self);
  GError *error = NULL;
  DBusGProxy *proxy;

  g_return_if_fail(priv != NULL);
  g_return_if_fail(priv->system_bus != NULL);

  if (priv->service_proxy != NULL || priv->connect_call != NULL)
    /* call already in progress */
    return;

  if (path == NULL || strlen(path) == 0) {
    g_warning("%s(): invalid path", __func__);
    return;
  }

  proxy = dbus_g_proxy_new_for_name_owner(priv->system_bus,
					  CONNMAN_SERVICE,
					  path,
					  CONNMAN_SERVICE_INTERFACE,
					  &error);

  if (error != NULL) {
    g_warning("failed to create dbus proxy for service %s: %s", path,
	      error->message);
    g_error_free(error);
    return;
  }

  if (proxy == NULL) {
    g_warning("failed to create dbus proxy for service %s", path);
    return;
  }

  priv->service_proxy = proxy;

  priv->connect_call = org_moblin_connman_Service_connect_async(proxy,
								connect_cb,
								self);
}

static void responded(GtkDialog *dialog, gint response_id, gpointer user_data)
{
  WirelessConnectDialog *self = WIRELESS_CONNECT_DIALOG(user_data);
  WirelessConnectDialogPriv *priv = WIRELESS_CONNECT_DIALOG_GET_PRIVATE(self);
  GtkTreeIter iter;
  gboolean result;
  gchar *path;

  gtk_widget_hide_all(GTK_WIDGET(self));

  if (response_id != GTK_RESPONSE_ACCEPT)
    return;

  /* connect to the service */

  result = gtk_combo_box_get_active_iter(GTK_COMBO_BOX(priv->network_combo),
					 &iter);
  if (!result) {
    path = gtk_combo_box_get_active_text(GTK_COMBO_BOX(priv->network_combo));

    /* FIXME: use ConnectService to connect to a hidden network */
    goto out;
  }

  gtk_tree_model_get(GTK_TREE_MODEL(priv->store), &iter,
		     COLUMN_PATH, &path, -1);
  connect(self, path);

out:
  g_free(path);
}

static void update_count_label(WirelessConnectDialog *self)
{
  WirelessConnectDialogPriv *priv = WIRELESS_CONNECT_DIALOG_GET_PRIVATE(self);
  gchar buf[200];

  g_snprintf(buf, sizeof(buf), _("%d networks detected"), priv->service_count);
  gtk_label_set_text(GTK_LABEL(priv->count_label), buf);
}

static void iterate_array(gpointer data, gpointer user_data)
{
  WirelessConnectDialog *self = WIRELESS_CONNECT_DIALOG(user_data);
  WirelessConnectDialogPriv *priv = WIRELESS_CONNECT_DIALOG_GET_PRIVATE(self);
  GValueArray *val_array = data;
  GHashTable *properties;
  const gchar *path, *name;
  GtkTreeIter iter;
  GValue *value;

  g_return_if_fail(val_array != NULL);

  if (val_array->n_values != 2)
    return;

  /* object path */
  value = g_value_array_get_nth(val_array, 0);
  g_return_if_fail(G_IS_VALUE(value));
  g_return_if_fail(G_VALUE_TYPE(value) == DBUS_TYPE_G_OBJECT_PATH);
  g_return_if_fail(G_VALUE_HOLDS_BOXED(value));

  path = g_value_get_boxed(value);
  if (!path)
    return;

  /* properties */
  value = g_value_array_get_nth(val_array, 1);
  g_return_if_fail(G_IS_VALUE(value));
  g_return_if_fail(G_VALUE_HOLDS_BOXED(value));

  properties = g_value_get_boxed(value);

  value = g_hash_table_lookup(properties, CONNMAN_PROPERTY_NAME);
  if (value == NULL)
    return;

  name = g_value_get_string(value);

  gtk_list_store_append(priv->store, &iter);
  gtk_list_store_set(priv->store, &iter, COLUMN_PATH, path,
		     COLUMN_NAME, name, -1);

  priv->service_count++;
}

static void get_services_cb(DBusGProxy *proxy, GPtrArray *array, GError *error,
		     gpointer user_data)
{
  WirelessConnectDialog *self;
  WirelessConnectDialogPriv *priv;

  g_return_if_fail(user_data != NULL);
  self =  WIRELESS_CONNECT_DIALOG(user_data);

  g_return_if_fail(self != NULL);
  g_return_if_fail(IS_WIRELESS_CONNECT_DIALOG(self));

  priv = WIRELESS_CONNECT_DIALOG_GET_PRIVATE(self);
  g_return_if_fail(priv != NULL);

  priv->services_call = NULL;

  if (error) {
    g_warning("%s(): failed to get properties from ofono manager: %s",
	      __func__, error->message);
    g_error_free(error);
    return;
  }

  gtk_list_store_clear(priv->store);
  priv->service_count = 0;

  g_ptr_array_foreach(array, iterate_array, self);
  g_ptr_array_free(array, TRUE);

  update_count_label(self);
}

static gboolean get_connman_manager(WirelessConnectDialog *self)
{
  WirelessConnectDialogPriv *priv = WIRELESS_CONNECT_DIALOG_GET_PRIVATE(self);

  if (priv->connman_manager != NULL)
    return TRUE;

  /* FIXME: handle connman missing or crashing */
  priv->connman_manager = dbus_g_proxy_new_for_name(priv->system_bus,
						    CONNMAN_SERVICE,
						    CONNMAN_MANAGER_PATH,
						    CONNMAN_MANAGER_INTERFACE);
  if (priv->connman_manager == NULL)
    return FALSE;

  return TRUE;
}

static void get_services(WirelessConnectDialog *self)
{
  WirelessConnectDialogPriv *priv = WIRELESS_CONNECT_DIALOG_GET_PRIVATE(self);
  DBusGProxyCall *call;

  if (priv->connman_manager == NULL)
    if (get_connman_manager(self) == FALSE)
      return;

  if (priv->services_call != NULL)
    /* call already in progress */
    return;

  call = org_moblin_connman_Manager_get_services_async(priv->connman_manager,
						       get_services_cb,
						       self);

  priv->services_call = call;
}

void wireless_connect_dialog_show(WirelessConnectDialog *self)
{
  g_return_if_fail(self != NULL);
  g_return_if_fail(IS_WIRELESS_CONNECT_DIALOG(self));

  gtk_widget_show_all(GTK_WIDGET(self));
  get_services(self);
}

WirelessConnectDialog *wireless_connect_dialog_new(void)
{
  GtkWidget *vbox, *hbox, *label;
  WirelessConnectDialogPriv *priv;
  WirelessConnectDialog *self;

  self = g_object_new(WIRELESS_CONNECT_DIALOG_TYPE, NULL);
  priv = WIRELESS_CONNECT_DIALOG_GET_PRIVATE(self);

  gtk_window_set_title(GTK_WINDOW(self), _("Connect to Wireless Network"));

  /* FIXME: pre-select connect button once user has made a selection */
  gtk_dialog_add_buttons(GTK_DIALOG(self),
			 GTK_STOCK_CANCEL, GTK_RESPONSE_REJECT,
			 GTK_STOCK_CONNECT, GTK_RESPONSE_ACCEPT,
			 NULL);

  vbox = gtk_dialog_get_content_area(GTK_DIALOG(self));

  hbox = gtk_hbox_new(FALSE, 5);
  gtk_box_pack_start(GTK_BOX(vbox), hbox, TRUE, TRUE, 5);

  label = gtk_label_new(_("Network name:"));
  gtk_box_pack_start(GTK_BOX(hbox), label, TRUE, TRUE, 5);

  priv->store = gtk_list_store_new(N_COLUMNS, G_TYPE_STRING, G_TYPE_STRING);
  priv->network_combo = gtk_combo_box_entry_new_with_model(GTK_TREE_MODEL(priv->store),
							   COLUMN_NAME);
  gtk_box_pack_start(GTK_BOX(hbox), priv->network_combo, TRUE, TRUE, 5);

  priv->count_label = gtk_label_new(_("0 networks detected"));
  update_count_label(self);
  gtk_box_pack_start(GTK_BOX(vbox), priv->count_label, TRUE, TRUE, 5);

  g_signal_connect(G_OBJECT(self), "response", G_CALLBACK(responded),
		   self);

  return self;
}
